"""
  ____       __                        __    __   __      _ __
 /_  / ___ _/ /  ___  ___ ___________ / /__ / /__/ /_____(_) /__
  / /_/ _ `/ _ \/ _ \/ -_) __/___/ -_) / -_)  '_/ __/ __/ /  '_/
 /___/\_,_/_//_/_//_/\__/_/      \__/_/\__/_/\_\\__/_/ /_/_/\_\

Copyright 2023 Zahner-Elektrik GmbH & Co. KG

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import numpy as np
import io
import struct
import datetime


def readF8FromFile(file: io.BufferedReader) -> float:
    bytesRead = file.read(8)
    retval = struct.unpack(">d", bytesRead)
    return retval[0]


def readF8ArrayFromFile(file: io.BufferedReader, length: int) -> np.ndarray:
    return np.ndarray(shape=(length,), dtype=">f8", buffer=file.read(8 * length))


def readI2FromFile(file: io.BufferedReader) -> int:
    return int.from_bytes(file.read(2), "big", signed=True)


def peekI2FromFile(file: io.BufferedReader) -> int:
    return int.from_bytes(file.peek(2), "big", signed=True)


def readI2ArrayFromFile(file: io.BufferedReader, length: int) -> np.ndarray:
    return np.ndarray(shape=(length,), dtype=">i2", buffer=file.read(2 * length))


def readI6FromFile(file: io.BufferedReader) -> int:
    return int.from_bytes(file.read(6), "big", signed=True)


def readTimeStampDateTimeArrayFromFile(
    file: io.BufferedReader, length: int
) -> list[datetime.datetime]:
    timeArray = readF8ArrayFromFile(file, length)
    datetimeArray = []
    for time in timeArray:
        datetimeArray.append(thalesTimeStampToDateTime(time))
    return datetimeArray


def readZahnerStringFromFile(file: io.BufferedReader) -> str:
    length = readI2FromFile(file)
    content = bytearray()

    for i in range(length):
        content.append(file.read(1)[0])

    return content.decode("ASCII").swapcase()


def readZahnerDate(file: io.BufferedReader) -> datetime.datetime:
    dateString = readZahnerStringFromFile(file)
    date = dateString[0:6]

    day = int(date[0:2])
    month = int(date[2:4])
    year = int(date[4:6])

    """
    Only the last two digits of the date are saved.
    It is assumed that the measurement was carried out between 1970 and 2070.
    A software update is necessary in the year 2070 at the latest.
    """
    if year < 70:
        year += 2000
    else:
        year += 1900

    return datetime.datetime(year, month, day)


def thalesTimeStampToDateTime(timestamp: int) -> datetime.datetime:
    """Calculation of the time stamp.

    The time is in seconds related to 01.01.1980.

    :param timestamp: Seconds since 01.01.1980.
    :returns: Python datetime object.
    """
    timeZero = datetime.datetime(1980, 1, 1)
    timeDifference = datetime.timedelta(seconds=abs(timestamp))

    timestamp = timeZero + timeDifference
    return timestamp



"""
  ____       __                        __    __   __      _ __
 /_  / ___ _/ /  ___  ___ ___________ / /__ / /__/ /_____(_) /__
  / /_/ _ `/ _ \/ _ \/ -_) __/___/ -_) / -_)  '_/ __/ __/ /  '_/
 /___/\_,_/_//_/_//_/\__/_/      \__/_/\__/_/\_\\__/_/ /_/_/\_\

Copyright 2023 Zahner-Elektrik GmbH & Co. KG

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


class ZahnerAnalysisError(Exception):
    """zahner_analysis exception class

    This exception is issued in case of an unspecific error.
    """

    message: str

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)
        return



"""
  ____       __                        __    __   __      _ __
 /_  / ___ _/ /  ___  ___ ___________ / /__ / /__/ /_____(_) /__
  / /_/ _ `/ _ \/ _ \/ -_) __/___/ -_) / -_)  '_/ __/ __/ /  '_/
 /___/\_,_/_//_/_//_/\__/_/      \__/_/\__/_/\_\\__/_/ /_/_/\_\

Copyright 2023 Zahner-Elektrik GmbH & Co. KG

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from zahner_analysis.file_import.ism_import import IsmImport
from zahner_analysis.file_import.impedance_model_import import (
    IsfxModelImport,
    IsfxModelElement,
    IsfxModelElementParameter,
)
from zahner_analysis.plotting.impedance_plot import bodePlotter, nyquistPlotter
from zahner_analysis.analysis_tools.analysis_connection import AnalysisConnection
from zahner_analysis.analysis_tools.error import ZahnerAnalysisError

import os
import json
import io
import logging
import time

DUMMY_MODEL = """<?xml version="1.0" encoding="UTF-8"?>
<zahner-impedance-model name="New Model">
    <zahner-fileinfo>
        <type>zahner-impedance-model</type>
        <file-version>3.0</file-version>
        <created>2023-04-13T07:51:22+02:00</created>
        <generator>Zahner Analysis</generator>
        <generator-version>3.3.5</generator-version>
        <comment></comment>
        <app-session>cef8c2e3-250d-47ff-9638-ef58933d107e</app-session>
        <file-id>e0e0c2a4-296f-40d0-9937-58cad6e47464</file-id>
    </zahner-fileinfo>
    <parsed-tree>
        <serial-connect>
            <resistor name="R0">
                <parameter index="0" value="100" fitterFixed="1"/>
            </resistor>
        </serial-connect>
    </parsed-tree>
    <schematics-graph>
        <node pos-x="0" type="1" name="R0" node-id="0" pos-y="-64">
            <parameter value="100" fitterFixed="1" index="0"/>
        </node>
    </schematics-graph>
</zahner-impedance-model>
"""


class EisFittingResult:
    """This class contains the results of the fit.

    The constructor of this class is called by the fit method.

    :param fitResult: Fit result as JSON string.
    :param fittedModel: Model with the fitted parameters.
    :param fittedSimulatedData: Data simulated from the model.
    :param fitInputData: Data used for the fit.
    """

    def __init__(
        self,
        fitResult: str,
        fittedModel: IsfxModelImport,
        fittedSimulatedData: IsmImport,
        fitInputData: IsmImport,
    ):
        self._fitResult = fitResult
        self._fittedModel = fittedModel
        self._fittedSimulatedData = fittedSimulatedData
        self._fitInputData = fitInputData
        return

    def getFitResultJson(self, fmt="json"):
        """Get the fit result.

        This function returns the fit result as JSON or as string.

        The following is an example of the data returned. For each parameter of each element the value,
        unit, significance and error is returned. Also data for the overall fit result is returned.

        .. code-block:: json

            {
              "model": {
                "C0": {
                  "C": {
                    "error": 15.151926309121153,
                    "significance": 0.032274504286208844,
                    "value": 0.05020110574526214,
                    "value_unit": "F"
                  }
                },
                "CPE0": {
                  "C_eq": {
                    "error": 1.055031648230169,
                    "significance": 0.2519304540206341,
                    "value": 0.06756234793290997,
                    "value_unit": "F"
                  },
                  "α": {
                    "error": 0.5970910562207303,
                    "significance": 0.6845012405311864,
                    "value": 0.7366826808067263,
                    "value_unit": ""
                  }
                },
                "FI0": {
                  "W": {
                    "error": 0.48845055246742036,
                    "significance": 0.6408552251285363,
                    "value": 0.04073380256581006,
                    "value_unit": "Ωs^(-½)"
                  },
                  "k": {
                    "error": 1.9687457791834184,
                    "significance": 0.15899771929695322,
                    "value": 0.0013854241183094132,
                    "value_unit": "1/s"
                  }
                },
                "L0": {
                  "L": {
                    "error": 2.8396295426843867,
                    "significance": 0.6911809622435452,
                    "value": 7.940027736136356e-07,
                    "value_unit": "H"
                  }
                },
                "R0": {
                  "R": {
                    "error": 1.5977793098280266,
                    "significance": 0.07859404364129204,
                    "value": 0.0036204263202214014,
                    "value_unit": "Ω"
                  }
                },
                "R1": {
                  "R": {
                    "error": 0.529684672149924,
                    "significance": 0.9232302468243536,
                    "value": 0.026830261706435942,
                    "value_unit": "Ω"
                  }
                },
                "R2": {
                  "R": {
                    "error": 0.18918999015755764,
                    "significance": 0.4215958599730566,
                    "value": 0.035941054916087775,
                    "value_unit": "Ω"
                  }
                }
              },
              "overall": {
                "impedance_error_max": 1.9436053656746468,
                "impedance_error_mean": 0.14339995178171594,
                "overall_error": 1.1323929422725485,
                "phase_error_max": 0.22001396123330905,
                "phase_error_mean": 0.018250258203149725
              }
            }

        :param fmt: "json" for json, else string.
        :returns: json or string.
        """
        if fmt == "json":
            return self._fitResult
        else:
            return json.dumps(
                self._fitResult, ensure_ascii=False, sort_keys=True, indent=2
            )

    def getFittedModel(self) -> IsfxModelImport:
        """Get the fitted model.

        :returns: The model.
        """
        return self._fittedModel

    def getFittedSimulatedData(self) -> IsmImport:
        """Get the fitted simulated data.

        These data were generated by a simulation of the model.

        :returns: The data.
        """
        return self._fittedSimulatedData

    def getFitInputData(self) -> IsmImport:
        """Get the samples used for fitting.

        These frequency points from the original data were used for the fit.
        These are the original measurement points used for the fit.
        These points were already smoothed or pre-processed with the ZHIT when it was set.

        :returns: The data.
        """
        return self._fitInputData

    def save(
        self,
        path="",
        foldername="",
        exist_ok=True,
        saveFitResultJson=True,
        saveFittedModel=True,
        saveFittedSimulatedSamples=True,
        saveFitInputSamples=True,
        fitResultJsonFilename="fit_result.json",
        fittedModelFilename="fitted.isfx",
        fittedSimulatedDataFilename="fitted_simulated.ism",
        fitInputDataFilename="fit_samples.ism",
    ):
        """Save all fit data.

        With this function, all the results files of the fit can be saved to the hard disk.
        It can be selected what all can be saved with. By default, everything is saved.

        :param path: Path where a folder with the fit results will be created.
            This path can also be relative, as with all paths in Python.
        :param foldername: Name of the folder in which the data will be saved.
        :param exist_ok: Parameter of the os.makedirs function which is used to create the folder.
            If exist_ok is False, an FileExistsError is raised if the target directory already exists.
        :param saveFitResultJson: If true, the json is stored under the filename of the variable fitResultJsonFilename.
        :param saveFittedModel: If true, the model is stored under the filename of the variable fittedModelFilename.
        :param saveFittedSimulatedSamples: If true, the simulated fitted data is stored under the filename of the variable fittedSimulatedDataFilename.
        :param saveFitInputSamples: If true, the samples used for fit are stored under the filename of the variable fitInputDataFilename.
        :param fitResultJsonFilename: Filename for the JSON with the fit result.
        :param fittedModelFilename: Filename for the fitted model.
        :param fittedSimulatedDataFilename: Filename for the data simulated with the model.
        :param fitInputDataFilename: Filename for the data of the points used for the fit.
        """
        path = os.path.join(path, foldername)
        os.makedirs(path, exist_ok=exist_ok)

        if saveFitResultJson:
            with open(
                os.path.join(path, fitResultJsonFilename), "w", encoding="utf-8"
            ) as f:
                f.write(self.getFitResultJson(fmt="txt"))

        if saveFittedModel:
            self._fittedModel.save(os.path.join(path, fittedModelFilename))

        if saveFittedSimulatedSamples:
            self._fittedSimulatedData.save(
                os.path.join(path, fittedSimulatedDataFilename)
            )

        if saveFitInputSamples:
            self._fitInputData.save(os.path.join(path, fitInputDataFilename))

        return

    def __str__(self):
        """Create string with informations about the model

        When this object is converted to a string, this function is called.

        :returns: A string with the information.
        """
        retval = f"Fitted Model:\n"
        retval += self.getFittedModel().toString()
        retval += "JSON fit result:\n"
        retval += self.getFitResultJson(fmt="txt")

        return retval


class EisFitting:
    """Class which can fit models to impedance spectra.

    This class uses the REST interface of the Zahner Analysis for fitting.


    This class uses the `Python logging module <https://docs.python.org/3/library/logging.html/>`_,
    which can be enabled and output with the following sample configuration.

    .. code-block:: python

        import logging

        if __name__ == "__main__":

            logging.basicConfig(
                filename='logfile.log',
                level=logging.DEBUG,
                format='%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s',
                datefmt='%Y-%m-%d %H:%M:%S',
            )

            conn = AnalysisConnection(
                ip="127.0.0.1",
                port=8085,
                tryToConnect=True,
                tryToStart=True,
                onlineCheckUrl="/id",
                apiKey="MyKeyToPreventSomeoneElseRemotelyUsingMyAnalysis",
                buffer_size=32768,
                keep_jobs=10,
                timeToWaitForOnline=10,
                keepOpen=True)

            fitting = ImpedanceFitting(conn)

    :param analysisConnection: Optional connection object to the Zahner Analysis.
        Not needed if Zahner Analysis is installed locally.
    """

    classAnalysisConnection: AnalysisConnection = None

    def __init__(self, analysisConnection: AnalysisConnection = None):
        self._analysisConnection = analysisConnection

        if self._analysisConnection is None:
            if EisFitting.classAnalysisConnection is None:
                EisFitting.classAnalysisConnection = AnalysisConnection()

            self._analysisConnection = EisFitting.classAnalysisConnection
        return

    def zhit(
        self,
        data: IsmImport = None,
        parameters: dict = {},
        timeout: float = None,
    ) -> IsmImport:
        """
        Performs the ZHIT evaluation.

        ZHIT is a software tool that uses measured phase data to reconstruct the impedance spectrum.
        The reconstructed impedance spectrum is then compared to the measured impedance spectrum to validate the measurement and identify any artifacts.

        Links to the topic ZHIT:
         * https://en.wikipedia.org/wiki/Z-HIT
         * https://doc.zahner.de/manuals/zahner_analysis.pdf

        Parameter dictionary for optional parameters:

        .. csv-table::
            :header-rows: 1

            Key , Description
            Smoothness , Factor with which smoothed. This must be determined empirically in the GUI.
            NumberOfSamples , Number of samples used for the data. Default all samples.

        .. code-block:: python

            parameters = {
                "Smoothness": 0.0002,
                "NumberOfSamples": 20
            }

        :param data: Data to which the ZHIT is applied.
        :param parameters: Optional parameters for the ZHIT.
        :param timeout: Timeout for the caculation.
        """
        model = IsfxModelImport(xmlString=DUMMY_MODEL)
        parameters["DataSource"] = "zhit"
        # A fit is performed and then everything is discarded except the data used for the fit.
        fitResult = self.fit(model, data, parameters, timeout=timeout)
        return fitResult.getFitInputData()

    def fit(
        self,
        model: IsfxModelImport,
        data: IsmImport = None,
        fitParams: dict = {},
        simulationParams: dict = {},
        timeout: float = None,
    ) -> EisFittingResult:
        """
        Performing the fit.

        With this method the model is fitted to the data.
        The initial values and the model can be easily developed using the Zahner Analysis GUI.

        The parameters are optional. You can also use only some optional parameters and omit others.
        Parameter dictionary for optional fit parameters:

        .. csv-table::
            :header-rows: 1

            Key , Description
            UpperFrequencyLimit , Upper frequency limit up to which the data is used for fitting. Default highest frequency.
            LowerFrequencyLimit , Lower frequency limit downto to which the data is used for fitting. Default lowest frequency.
            DataSource , Selection of the data to be used for the fit. "original" or "smoothed" or "zhit". Default "original".
            Smoothness , Factor with which smoothed. This must be determined empirically in the GUI.
            NumberOfSamples , Number of samples used for the fit. Default all samples.

        .. code-block:: python

            fitParams = {
                "UpperFrequencyLimit": 100000,
                "LowerFrequencyLimit": 1e-6,
                "DataSource": "zhit",  # "original", "smoothed" or "zhit"
                "Smoothness": 0.0002,
                "NumberOfSamples": 20
            }

        Parameter dictionary for optional simulation parameters:

        .. csv-table::
            :header-rows: 1

            Key , Description
            UpperFrequencyLimit , Upper frequency limit up to which the model is simulated. Default highest frequency of data.
            LowerFrequencyLimit , Lower frequency limit downto to which the model is simulated. Default lowest frequency of data.
            NumberOfSamples , Number of samples used for the simulation. Default 100.

        .. code-block:: python

            simulationParams = {
                "UpperFrequencyLimit": 10e3,
                "LowerFrequencyLimit": 1e-6,
                "NumberOfSamples": 150
            }

        :param model: Model which is fitted to the data.
        :param data: Data to which the model is fitted.
        :param fitParams: Explained in the previous text.
        :param simulationParams: Explained in the previous text.
        :param timeout: Timeout for the fit.
        """

        paramsDict = dict()
        paramsDict["job"] = "EvalEis.Fit"
        paramsDict["parameters"] = {"Fit": fitParams, "Simulation": simulationParams}
        paramsDict["mode"] = "queued"

        jobId = self._startFit(model, data, paramsDict)
        logging.debug("Zahner Analysis job-id: " + jobId)
        fitResult = self._waitForJob(jobId, timeout)
        if fitResult is None:
            raise ZahnerAnalysisError("Operation in Zahner Analysis Server failed")
        fittedModel = self._readFittedModel(jobId)
        fittedSimulatedSamples = self._readSimulatedData(jobId)
        fitInputSamples = self._readFitInputData(jobId)

        return EisFittingResult(
            fitResult, fittedModel, fittedSimulatedSamples, fitInputSamples
        )

    def _startFit(self, model: IsfxModelImport, data: IsmImport, params: dict) -> str:
        """
        Function which starts the fit.

        This function sends the model, data and parameters to the Zahner Analysis via HTTP post request.
        The JobId assigned by Zahner Analysis is returned.

        :param model: Model which is fitted to the data.
        :param data: Data to which the model is fitted.
        :param dict: Dictionary {"Fit":fitParams, "Simulation":simulationParams}.
        :returns: JobId
        """
        files = [
            ("eis-file", (data.getFileName(), io.BytesIO(data.getBinaryFileContent()))),
            (
                "model-file",
                (model.getFileName(), io.BytesIO(model.getBinaryFileContent())),
            ),
            ("job", (None, json.dumps(params).encode("utf-8"))),
        ]
        reply = self._analysisConnection.post("/job/start", files=files)

        if reply.status_code == 200:
            replyContent = json.loads(reply.content)
            jobId = replyContent["job-id"]

            if replyContent["status"] in ["failed", "error"]:
                logging.error("Zahner Analysis reply: " + str(replyContent))
        else:
            logging.error("Zahner Analysis reply: " + str(replyContent))

        return jobId

    def simulate(
        self, model: IsfxModelImport, simulationParams: dict, timeout: float = None
    ) -> IsmImport:
        """
        Simulate the model.

        With this method, an impedance spectrum is generated from the model.

        Parameter dictionary for simulation parameters:

        .. csv-table::
            :header-rows: 1

            Key , Description
            UpperFrequencyLimit , Upper frequency limit up to which the model is simulated. Default highest frequency of data.
            LowerFrequencyLimit , Lower frequency limit downto to which the model is simulated. Default lowest frequency of data.
            NumberOfSamples , Number of samples used for the simulation. Default 100.

        .. code-block:: python

            simulationParams = {
                "UpperFrequencyLimit": 10e3,
                "LowerFrequencyLimit": 1e-6,
                "NumberOfSamples": 150
            }

        :param model: Model to be simulated.
        :param simulationParams: Explained in the previous text.
        :param timeout: Timeout for the fit.
        """

        paramsDict = dict()
        paramsDict["job"] = "EvalEis.Sim"
        paramsDict["parameters"] = {"Simulation": simulationParams}
        paramsDict["mode"] = "queued"

        jobId = self._startSimulation(model, paramsDict)
        logging.debug("Zahner Analysis job-id: " + jobId)
        self._waitForJob(jobId, timeout)

        return self._readSimulatedData(jobId)

    def _startSimulation(self, model: IsfxModelImport, params: dict) -> str:
        """
        Function which starts the simulation.

        This function sends the model and parameters to the Zahner Analysis via HTTP post request.
        The JobId assigned by Zahner Analysis is returned.

        :param model: Model to be simulated.
        :param dict: Dictionary {"Simulation":simulationParams}.
        :returns: JobId
        """
        files = [
            (
                "model-file",
                (model.getFileName(), io.BytesIO(model.getBinaryFileContent())),
            ),
            ("job", (None, json.dumps(params).encode("utf-8"))),
        ]
        reply = self._analysisConnection.post("/job/start", files=files)

        if reply.status_code == 200:
            replyContent = json.loads(reply.content)
            jobId = replyContent.get("job-id")

            if replyContent["status"] in ["failed", "error"]:
                logging.error("Zahner Analysis reply: " + str(replyContent))
        else:
            logging.error("Zahner Analysis reply: " + str(replyContent))

        return jobId

    def _waitForJob(self, jobId: str, timeout: float = None):
        """
        Function which is waiting for the fit result.

        This function polls the status of the fit operation and reads the fit result as JSON.

        :param jobId: JobId.
        :param timeout: Time for which polling should be done in seconds, or None for infinite.
        :returns: Fit result as JSON string.
        """
        result = None
        continueWait = True
        startTime = time.time()

        while continueWait:
            reply = self._analysisConnection.get(f"/job/{jobId}/status")

            if reply.status_code == 200:
                replyContent = json.loads(reply.content)
                jobStatus = replyContent["status"]

                if jobStatus == "done":
                    continueWait = False
                    result = replyContent.get("result")
                elif jobStatus == "failed":
                    continueWait = False
                    logging.error("Zahner Analysis fitting failed")
                else:
                    diffTime = time.time() - startTime
                    if timeout is not None:
                        if diffTime > timeout:
                            logging.error("Zahner Analysis fitting timeout")
                            continueWait = False
                    if continueWait == True:
                        time.sleep(0.02)  # poll status every 20 ms

            else:
                continueWait = False
                logging.error(
                    "Zahner Analysis reply: " + str(json.loads(reply.content))
                )

        return result

    def _readFittedModel(self, jobId) -> IsfxModelImport:
        """
        Reading the fitted model from the Zahner Analysis.

        Reading is done via http get request.

        :param jobId: JobId.
        :returns: Object from received data.
        """
        reply = self._analysisConnection.get(f"/job/{jobId}/model")
        return IsfxModelImport(xmlString=reply.content.decode("utf-8"))

    def _readFitInputData(self, jobId) -> IsmImport:
        """
        Reading the Samples used for fit from the Zahner Analysis.

        Reading is done via http get request.

        :param jobId: JobId.
        :returns: Object from received data.
        """
        reply = self._analysisConnection.get(f"/job/{jobId}/samples")
        return IsmImport(reply.content)

    def _readSimulatedData(self, jobId) -> IsmImport:
        """Reading the simulation data from the Zahner Analysis.

        Reading is done via http get request.

        :param jobId: JobId.
        :returns: Object from received data.
        """
        reply = self._analysisConnection.get(f"/job/{jobId}/simulation")
        return IsmImport(reply.content)


class EisFittingPlotter:
    """
    Class with utility Nyquist and Bode plotting methods.

    This class contains methods to display the fit results in the style of Zahner Analysis.
    """

    @staticmethod
    def plotBode(
        fittingResult: EisFittingResult,
        impedanceData: IsmImport = None,
        axes=None,
        zTogetherPhase=True,
        absPhase=True,
    ):
        """Plotting the data in the Bode plot.

        For plotting matplotlib is used with the function :meth:`zahner_analysis.plotting.impedance_plot.bodePlot`.
        With this function or also only with matplotlib the plot can be represented adapted.
        This method displays the plot in the standard Zahner Analysis design.

        Either axes can be passed on which will be plotted, or a new figure with axes will be created
        automatically. The figure and the axes are always returned.

        If the impedanceData is not passed, then the samples used for the fit are displayed.

        The Zahner Analysis default setting impedance and phase in a plot and phase in magnitude can be deactivated with two parameters.

        The following code block shows a few snippets as an example:

        .. code-block:: python

            fig, ax = plt.subplots(2,2)

            (fig, (impedanceAxis, phaseAxis)) = EisFittingPlotter.plotBode(fittingResult, axes=(ax[0,0],ax[1,0]))
            (fig, (impedanceAxis, phaseAxis)) = EisFittingPlotter.plotBode(fittingResult, axes=(ax[0,1],ax[1,1]))
            impedanceAxis.legend(["Measured Data", "Fitted Model"])
            plt.show()

            # or

            (fig2, (impedanceAxis2, phaseAxis2)) = EisFittingPlotter.plotBode(fittingResult)
            impedanceAxis2.legend(["Measured Data", "Fitted Model"])
            fig2.set_size_inches(18, 10)
            plt.show()


        :param fittingResult: Fitting result object.
        :param impedanceData: Optional impedance data for display.
        :param axes: Tuple (impedanceAxis, phaseAxis) with impedance and phase axes object, or None if a new figure should be created.
        :param zTogetherPhase: Default True to display phase and impedance in one plot.
        :param absPhase: Default True to plot the absolute value of the phase.
        :returns: Tuple fig, axes
        """
        fig = None

        if impedanceData is not None:
            (fig, axes) = bodePlotter(
                axes,
                impedanceObject=impedanceData,
                zTogetherPhase=zTogetherPhase,
                absPhase=absPhase,
            )
        else:
            (fig, axes) = bodePlotter(
                axes,
                impedanceObject=fittingResult.getFitInputData(),
                zTogetherPhase=zTogetherPhase,
                absPhase=absPhase,
            )

        (fig, axes) = bodePlotter(
            axes,
            impedanceObject=fittingResult.getFittedSimulatedData(),
            zTogetherPhase=zTogetherPhase,
            absPhase=absPhase,
            argsImpedanceAxis={"linestyle": "solid", "marker": None},
            argsPhaseAxis={"linestyle": "solid", "marker": None},
        )

        return fig, axes

    @staticmethod
    def plotNyquist(
        fittingResult: EisFittingResult,
        impedanceData: IsmImport = None,
        ax=None,
        minusNyquist=True,
        maximumAbsImpedance: float = None,
    ):
        """
        Plotting the data in the Nyquist plot.

        For plotting matplotlib is used with the function :meth:`zahner_analysis.plotting.impedance_plot.nyquistPlot`.
        With this function or also only with matplotlib the plot can be represented adapted.
        This method displays the plot in the standard Zahner Analysis design.

        Either axes can be passed on which will be plotted, or a new figure with axes will be created
        automatically. The figure and the axes are always returned.

        If the impedanceData is not passed, then the samples used for the fit are displayed.

        The following code block shows a snippet as an example:

        .. code-block:: python

            (fig, ax) = EisFittingPlotter.plotBode(fittingResult, maximumAbsImpedance=1)
            ax.legend(["Measured Data", "Fitted Model"])
            plt.show()

        :param fittingResult: Fitting result object.
        :param impedanceData: Optional impedance data for display.
        :param ax: The axis on which to plot, or None to create a new figure.
        :param minusNyquist: Default True to invert the imaginary part of the impedance.
        :param maximumAbsImpedance: If the value is not None, only impedances whose absolute value is smaller than this value are plotted.
        :returns: Tuple fig, ax
        """
        fig = None

        if impedanceData is not None:
            (fig, ax) = nyquistPlotter(
                ax,
                impedanceObject=impedanceData,
                minusNyquist=minusNyquist,
                maximumAbsImpedance=maximumAbsImpedance,
            )
        else:
            (fig, ax) = nyquistPlotter(
                ax,
                impedanceObject=fittingResult.getFitInputData(),
                minusNyquist=minusNyquist,
                maximumAbsImpedance=maximumAbsImpedance,
            )

        (fig, ax) = nyquistPlotter(
            ax,
            impedanceObject=fittingResult.getFittedSimulatedData(),
            minusNyquist=minusNyquist,
            argsNyquistAxis={"linestyle": "solid", "marker": None, "color": "#40e0d0"},
            maximumAbsImpedance=maximumAbsImpedance,
        )

        return fig, ax



from zahner_analysis.analysis_tools.eis_fitting import EisFitting
from zahner_analysis.file_import.impedance_model_import import IsfxModelImport
from zahner_analysis.file_import.ism_import import IsmImport
from zahner_analysis.plotting.impedance_plot import bodePlotter
import matplotlib.pyplot as plt

if __name__ == "__main__":
    fitting = EisFitting()

    dataWithDrift = IsmImport("drift.ism")
    dataWithZHIT = fitting.zhit(dataWithDrift)

    (fig1, (impedanceAxis1, phaseAxis1)) = bodePlotter(impedanceObject=dataWithDrift)
    (fig1, (impedanceAxis1, phaseAxis1)) = bodePlotter(
        (impedanceAxis1, phaseAxis1),
        impedanceObject=dataWithZHIT,
        argsImpedanceAxis={"linestyle": "solid", "marker": None},
        argsPhaseAxis={"linestyle": "solid", "marker": None},
    )
    phaseAxis1.legend(["Original", "ZHIT"])
    fig1.set_size_inches(18, 10)

    fitParams = {
        "DataSource": "zhit",
        "Smoothness": 0.0002,
        "NumberOfSamples": 40,
    }

    impedanceCircuitModel = IsfxModelImport("RC-R-L.isfx")
    fittingResult = fitting.fit(
        impedanceCircuitModel, dataWithDrift, fitParams=fitParams
    )
    print(fittingResult)

    (fig2, (impedanceAxis2, phaseAxis2)) = bodePlotter(impedanceObject=dataWithDrift)
    (fig2, (impedanceAxis2, phaseAxis2)) = bodePlotter(
        (impedanceAxis2, phaseAxis2),
        impedanceObject=fittingResult.getFittedSimulatedData(),
        argsImpedanceAxis={"linestyle": "solid", "marker": None},
        argsPhaseAxis={"linestyle": "solid", "marker": None},
    )
    (fig2, (impedanceAxis2, phaseAxis2)) = bodePlotter(
        (impedanceAxis2, phaseAxis2),
        impedanceObject=fittingResult.getFitInputData(),
        argsImpedanceAxis={"linestyle": "None", "marker": "x"},
        argsPhaseAxis={"linestyle": "None", "marker": "x"},
    )

    phaseAxis2.legend(["Original", "Fitted", "ZHIT"])
    fig2.set_size_inches(18, 10)

    fittingResult.getFitInputData().save("ZHIT.ism")

    plt.show()



from zahner_analysis.analysis_tools.setup_compensation import SetupCompensation
from zahner_analysis.plotting.impedance_plot import bodePlotter
from zahner_analysis.file_import.ism_import import IsmImport
from zahner_analysis.file_export.ism_export import IsmExport
import matplotlib.pyplot as plt

if __name__ == "__main__":
    originalData = IsmImport("./ExampleData/500uR_measurement_pro.ism")

    compensatedDataFileName = (
        "./ExampleData/500uR_measurement_pro_short_compensated.ism"
    )

    compensationData = SetupCompensation(
        shortData="./ExampleData/short_circuit_measurement_pro.ism",
        openData=None,
        loadData=None,
        referenceData=None,
    )

    compensationData.setSmoothingWindowLength(3)
    compensationData.setSmoothingPolyOrder(2)

    compensatedData = compensationData.compensateIsm(originalData)

    (fig1, (impedanceAxis1, phaseAxis1)) = bodePlotter(
        impedanceObject=originalData,
        zTogetherPhase=False,
        absPhase=False,
        argsImpedanceAxis={"marker": None},
        argsPhaseAxis={"marker": None},
    )

    (fig1, (impedanceAxis1, phaseAxis1)) = bodePlotter(
        (impedanceAxis1, phaseAxis1),
        impedanceObject=compensatedData,
        zTogetherPhase=False,
        absPhase=False,
        argsImpedanceAxis={"linestyle": "solid", "marker": None},
        argsPhaseAxis={"linestyle": "solid", "marker": None},
    )

    impedanceAxis1.legend(["Measured Data", "Compensated Data"])
    phaseAxis1.legend(["Measured Data", "Compensated Data"])

    impedanceAxis1.set_ylim([250e-6, 1e-3])
    phaseAxis1.set_ylim([-15.0, +45.0])

    fig1.set_size_inches(18, 10)
    plt.show()

    exportFile = IsmExport(compensatedData)
    exportFile.writeToFile(compensatedDataFileName)



from zahner_analysis.analysis_tools.eis_fitting import EisFitting, EisFittingPlotter
from zahner_analysis.file_import.impedance_model_import import IsfxModelImport
from zahner_analysis.file_import.ism_import import IsmImport
from zahner_analysis.analysis_tools.analysis_connection import AnalysisConnection
from zahner_analysis.plotting.impedance_plot import bodePlotter
import matplotlib.pyplot as plt

if __name__ == "__main__":
    conn = AnalysisConnection(
        ip="127.0.0.1",
        port=8085,
        tryToConnect=True,
        tryToStart=True,
        onlineCheckUrl="/id",
        apiKey="MyKeyToPreventSomeoneElseRemotelyUsingMyAnalysis",
        buffer_size=32768,
        keep_jobs=10,
        timeToWaitForOnline=10,
        keepOpen=True,
    )

    fitting = EisFitting(conn)

    fitParams = {
        "UpperFrequencyLimit": 100000,
        "LowerFrequencyLimit": 1e-6,
        "DataSource": "zhit",  # "original", "smoothed" or "zhit"
        "Smoothness": 0.0002,
        "NumberOfSamples": 20,
    }

    simulationParams = {
        "UpperFrequencyLimit": 10e3,
        "LowerFrequencyLimit": 1e-6,
        "NumberOfSamples": 150,
    }

    impedanceCircuitModel = IsfxModelImport("li-ion-model.isfx")
    impedanceData = IsmImport("li-ion-battery.ism")

    simulatedData = fitting.simulate(impedanceCircuitModel, simulationParams)

    (fig1, (impedanceAxis1, phaseAxis1)) = bodePlotter(impedanceObject=impedanceData)
    (fig1, (impedanceAxis1, phaseAxis1)) = bodePlotter(
        (impedanceAxis1, phaseAxis1),
        impedanceObject=simulatedData,
        argsImpedanceAxis={"linestyle": "solid", "marker": None},
        argsPhaseAxis={"linestyle": "solid", "marker": None},
    )
    phaseAxis1.legend(["Measured Data", "Simulated Model with start values"])
    fig1.set_size_inches(18, 10)
    plt.show()

    foldername = "fit_result"

    fig1.savefig(f"{foldername}/bode_not_fitted.png")

    fittingResult = fitting.fit(
        impedanceCircuitModel,
        impedanceData,
        fitParams=fitParams,
        simulationParams=simulationParams,
    )

    print(fittingResult)

    fittingResult.save(
        path="",
        foldername=foldername,
        exist_ok=True,
        saveFitResultJson=True,
        saveFittedModel=True,
        saveFittedSimulatedSamples=True,
        saveFitInputSamples=True,
        fitResultJsonFilename="fit_result.json",
        fittedModelFilename="fitted.isfx",
        fittedSimulatedDataFilename="fitted_simulated.ism",
        fitInputDataFilename="fit_samples.ism",
    )

    (fig2, (impedanceAxis2, phaseAxis2)) = EisFittingPlotter.plotBode(
        fittingResult, impedanceData
    )
    impedanceAxis2.legend(
        impedanceAxis2.get_lines() + phaseAxis2.get_lines(),
        2 * ["Measured Data", "Fitted Model"],
    )
    fig2.set_size_inches(18, 10)
    plt.show()
    fig2.savefig(f"{foldername}/bode_fitted.pdf")

    (fig3, (impedanceAxis3, phaseAxis3)) = EisFittingPlotter.plotBode(
        fittingResult, impedanceData, zTogetherPhase=False, absPhase=False
    )
    impedanceAxis3.legend(["Measured Data", "Fitted Model"])
    phaseAxis3.legend(["Measured Data", "Fitted Model"])
    fig3.set_size_inches(18, 10)
    plt.show()
    fig3.savefig(f"{foldername}/bode.svg")

    (fig4, ax) = EisFittingPlotter.plotNyquist(
        fittingResult, impedanceData, minusNyquist=False, maximumAbsImpedance=0.8
    )
    ax.legend(["Measured Data", "Fitted Model"])
    fig4.set_size_inches(15, 15)
    plt.show()

    fig4.savefig(f"{foldername}/nyquist.jpg")

    simulated = fittingResult.getFittedSimulatedData()

    fig5, (axis) = plt.subplots(1, 1)
    axis.loglog(
        simulated.getFrequencyArray(),
        simulated.getImpedanceArray(),
        marker="o",
        linewidth=0,
        fillstyle="none",
    )
    axis.loglog(
        impedanceData.getFrequencyArray(),
        impedanceData.getImpedanceArray(),
        marker="x",
        linewidth=0,
        fillstyle="none",
    )
    axis.grid("both")
    axis.set_xlabel(r"f")
    axis.set_ylabel(r"|Z|")
    fig5.set_size_inches(18, 10)
    plt.show()



from zahner_analysis.file_import.ism_import import IsmImport
from scipy.signal import savgol_filter
import numpy as np
import copy
from typing import Union, Optional


def interpolateIsmToImpedanceArray(
    ism: IsmImport, frequencyPoints: list[float]
) -> np.array:
    """
    Function for the interpolation of spectra.

    :param ism: Object with spectrum.
    :param frequencyPoints: Frequency points which are to be interpolated.
    :return: Array with interpolated impedance points.
    """
    return np.interp(
        frequencyPoints, ism.getFrequencyArray(), ism.getComplexImpedanceArray()
    )


def smoothIsm(ism: IsmImport, window_length: int, polyorder: int) -> IsmImport:
    """
    Smoothen the ism data with a savgol filter.

    https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.savgol_filter.html

    :param ism: ism data which are smoothed.
    :param window_length: Window length for smoothening.
    :param polyorder: Polynomial order must be greater than window length.
    :return: Smoothed ism data.
    """
    freqArray = ism.getFrequencyArray()
    ism.frequency = freqArray
    ism.impedance = savgol_filter(ism.getImpedanceArray(), window_length, polyorder)
    ism.phase = savgol_filter(ism.getPhaseArray(), window_length, polyorder)
    ism.fromIndex = 0
    ism.toIndex = len(ism.impedance)
    ism.swapNecessary = False
    return ism


class SetupCompensation(object):
    """
    Object which contains the data and methods to compensate ism data.

    ⚠️WARNING⚠️

    * **The results must be checked for plausibility, since incorrectly performed short, open and load measurements can also degrade the measurement results.**

    * **With a second measurement of a known object it must be verified if the compensation works or if it worsens the result.**

    * **The cable routing must not change after the calibration measurement. Since the cable routing and its impedance and parasitic properties are part of the calibration data.**

    * **For the best possible correction, the point density of the calibration measurements must be greater than or equal to that of the measurement to be corrected.**

    * **The calibration data are smoothed before the calculation, for this the window length and the polynomial order must be defined.**

    * **The order of the polynomial and the length of the window must be set by the user, this affects the result.**

    The formulas for the correction were derived from a theoretical four terminal network, which corresponds to the parasitic behavior of the measuring device and the setup.
    This formula is universal. It does not depend on the instrument used:

    .. math::

        Z=\\frac{(Z_{\\text{short}}-Z_{\\text{meas}})(Z_{\\text{load}}-Z_{\\text{open}})}{(Z_{\\text{short}}-Z_{\\text{load}})(Z_{\\text{meas}}-Z_{\\text{open}})} \cdot Z_{\\text{load,ref}}

    .. list-table::
        :widths: 50 50
        :header-rows: 1

        * - Parameter
          - Description
        * - :math:`Z_{\\text{short}}`
          - Measurement without object. Short-circuited 4-wire test setup. CE, RE, WE power and WE sense are connected together.
        * - :math:`Z_{\\text{open}}`
          - Measurement without object. CE and RE as well as WE power and WE sense are connected to each other.
        * - :math:`Z_{\\text{load}}`
          - Measurement with a reference object of known defined impedance over the frequency range.
        * - :math:`Z_{\\text{load,ref}}`
          - Real impedance of the reference object of $Z_{\text{load}}$ measurement.
        * - :math:`Z_{\\text{meas}}`
          - Measurement to be corrected.
        * - :math:`Z`
          - Measurement result corrected with the short, open and load data.

    Basic Usage:

    .. code-block::


        originalData = IsmImport("./ExampleData/500uR_measurement_pro.ism")

        compensatedDataFileName = (
            "./ExampleData/500uR_measurement_pro_short_compensated.ism"
        )

        compensationData = SetupCompensation(
            shortData="./ExampleData/short_circuit_measurement_pro.ism"
        )

        compensatedData = compensationData.compensateIsm(originalData)

    :param shortData: short data for correction, defaults to None.
    :param openData: open data for correction, defaults to None.
    :param loadData: load data for correction, defaults to None.
    :param referenceData: real value of load data for correction, defaults to None. Here you can simply pass a number, for example 1 for 1 Ohm reference object.
    :param smoothingWindowLength: Length of the smoothing window, defaults to 5.
    :param smoothingPolyOrder: Length of the smoothing poly, defaults to 3.
    """

    def __init__(
        self,
        shortData: Optional[Union[IsmImport, str, complex]] = None,
        openData: Optional[Union[IsmImport, str, complex]] = None,
        loadData: Optional[Union[IsmImport, str, complex]] = None,
        referenceData: Optional[Union[IsmImport, str, complex]] = None,
        smoothingWindowLength: Optional[int] = 5,
        smoothingPolyOrder: Optional[int] = 3,
    ):
        if shortData is None:
            self.shortData = 0 + 0j
        else:
            if isinstance(shortData, str):
                self.shortData = IsmImport(shortData)
            else:
                self.shortData = shortData

        if openData is None:
            self.openData = 1e15 + 0j
        else:
            if isinstance(openData, str):
                self.openData = IsmImport(openData)
            else:
                self.openData = openData

        if loadData is None and referenceData is None:
            self.loadData = 1 + 0j
            self.referenceData = 1 + 0j

        elif loadData is not None and referenceData is not None:
            if isinstance(loadData, str):
                self.loadData = IsmImport(loadData)
            else:
                self.loadData = loadData

            if isinstance(referenceData, str):
                self.referenceData = IsmImport(referenceData)
            elif isinstance(referenceData, IsmImport):
                self.referenceData = referenceData
            else:
                self.referenceData = referenceData
        else:
            raise ValueError("loadData requiers referenceData")

        self.smoothingWindowLength = smoothingWindowLength
        self.smoothingPolyOrder = smoothingPolyOrder
        return

    def setSmoothingWindowLength(self, value: int):
        """
        Sets the length of the smoothing window. Must be smaller than the poly length.

        :param value: Length of the smoothing window.
        """
        self.smoothingWindowLength = value
        return

    def getSmoothingWindowLength(self) -> int:
        """
        Length of the smoothing window.

        :return: Length of the smoothing window.
        """
        return self.smoothingWindowLength

    def setSmoothingPolyOrder(self, value: int):
        """
        Sets the length of the smoothing polynomial. Must be greater than the window length.

        :param value: Length of the smoothing poly.
        """
        self.smoothingPolyOrder = value
        return

    def getSmoothingPolyOrder(self) -> int:
        """
        Length of the smoothing poly.

        :return: Length of the smoothing poly.
        """
        return self.smoothingPolyOrder

    def compensateIsm(
        self, ism: Union[IsmImport, str], conjugateShort: bool = False
    ) -> IsmImport:
        """
        Compensate ism Data.

        This function can be used to apply the Short Open and Load data to the passed impedance data.
        You can then call this function several times to correct different data with the same data.

        :param ism: Data which are to be compensated.
        :param conjugateShort: Conjugate the short data., defaults to False.
        :return: Impedance object which was compensated.
        """
        if isinstance(ism, str):
            ism = IsmImport(ism)

        Zism = ism.getComplexImpedanceArray()

        if isinstance(self.shortData, IsmImport):
            Zshort = interpolateIsmToImpedanceArray(
                smoothIsm(
                    self.shortData, self.smoothingWindowLength, self.smoothingPolyOrder
                ),
                ism.getFrequencyArray(),
            )
        else:
            Zshort = np.array([self.shortData for _ in ism.getFrequencyArray()])

        if isinstance(self.openData, IsmImport):
            Zopen = interpolateIsmToImpedanceArray(
                smoothIsm(
                    self.openData, self.smoothingWindowLength, self.smoothingPolyOrder
                ),
                ism.getFrequencyArray(),
            )
        else:
            Zopen = np.array([self.openData for _ in ism.getFrequencyArray()])

        if isinstance(self.loadData, IsmImport):
            Zload = interpolateIsmToImpedanceArray(
                smoothIsm(
                    self.loadData, self.smoothingWindowLength, self.smoothingPolyOrder
                ),
                ism.getFrequencyArray(),
            )
        else:
            Zload = np.array([self.loadData for _ in ism.getFrequencyArray()])

        if isinstance(self.referenceData, IsmImport):
            Zreference = interpolateIsmToImpedanceArray(
                smoothIsm(
                    self.referenceData,
                    self.smoothingWindowLength,
                    self.smoothingPolyOrder,
                ),
                ism.getFrequencyArray(),
            )
        else:
            Zreference = np.array([self.referenceData for _ in ism.getFrequencyArray()])

        if conjugateShort == True:
            Zshort = np.conjugate(Zshort)

        Zcompensated = (
            ((Zshort - Zism) * (Zload - Zopen)) / ((Zism - Zopen) * (Zshort - Zload))
        ) * Zreference

        compensatedIsm = copy.deepcopy(ism)
        compensatedIsm.frequency = ism.getFrequencyArray()
        compensatedIsm.impedance = np.abs(Zcompensated)
        compensatedIsm.phase = np.angle(Zcompensated)
        compensatedIsm.fromIndex = 0
        compensatedIsm.toIndex = len(compensatedIsm.impedance)
        compensatedIsm.swapNecessary = False

        return compensatedIsm



from zahner_analysis.analysis_tools.eis_fitting import EisFitting
from zahner_analysis.file_import.impedance_model_import import IsfxModelImport
from zahner_analysis.file_import.ism_import import IsmImport
import matplotlib.pyplot as plt
from matplotlib.ticker import EngFormatter

import os
import glob
import re

if __name__ == "__main__":
    impedanceCircuitModel = IsfxModelImport(r"diode-ac-model.isfx")
    print(impedanceCircuitModel)

    print(f"Resistance initial value: {impedanceCircuitModel['R0']['R'].getValue()}\n")

    fitting = EisFitting()

    spectras = glob.glob(os.path.join("measured_spectras", "*_mvdc.ism"))

    measuredVoltageValues = []
    fittedResistanceValues = []

    for spectra in spectras:
        # Fitting the spectra in a loop
        print(f"File: {spectra}")
        fittingResult = fitting.fit(impedanceCircuitModel, IsmImport(spectra))

        # Append arrys with fit results
        voltagePattern = r"(\d+)_mvdc.ism"
        voltage = float(re.search(voltagePattern, spectra)[1]) / 1000.0
        measuredVoltageValues.append(voltage)
        fittedResistanceValue = fittingResult.getFittedModel()["R0"]["R"].getValue()
        fittedResistanceValues.append(fittedResistanceValue)

        voltageFormatter = EngFormatter(unit="V", sep=" ")
        resistanceFormatter = EngFormatter(unit="Ω", sep=" ")
        print(
            f"Voltage: {voltageFormatter.format_data(voltage)} Fitted Resistance: {resistanceFormatter.format_data(fittedResistanceValue)}"
        )

        # Save the fit results
        filename = os.path.splitext(os.path.split(spectra)[1])[0]
        fittingResult.save(path="fitted_spectras", foldername=filename)

        # Setting the new model for the next fit
        impedanceCircuitModel = fittingResult.getFittedModel()

    fig, (ax) = plt.subplots(1, 1)

    ax.semilogy(measuredVoltageValues, fittedResistanceValues)
    ax.xaxis.set_major_formatter(EngFormatter(unit="V"))
    ax.yaxis.set_major_formatter(EngFormatter(unit="Ω"))
    ax.set_xlabel(r"Voltage across the diode")
    ax.set_ylabel(r"Differential Resistance")
    ax.grid(which="both", linestyle="--", linewidth=0.5)
    fig.set_size_inches(12, 10)
    fig.savefig("RvsU.svg")
    plt.show()



"""
  ____       __                        __    __   __      _ __
 /_  / ___ _/ /  ___  ___ ___________ / /__ / /__/ /_____(_) /__
  / /_/ _ `/ _ \/ _ \/ -_) __/___/ -_) / -_)  '_/ __/ __/ /  '_/
 /___/\_,_/_//_/_//_/\__/_/      \__/_/\__/_/\_\\__/_/ /_/_/\_\

Copyright 2023 Zahner-Elektrik GmbH & Co. KG

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import numpy as np
import datetime
import io
import os
import datetime
from typing import Union

from zahner_analysis.file_import.thales_file_utils import *


class IsmImport:
    """
    Class to be able to read ism files.

    This class extracts the data from the ism files.
    It returns the data for the frequency range between the reversal frequency and the end frequency.

    :param file: The path to the ism file, or the ism file as bytes or bytearray.
    """

    def __init__(self, filename: Union[str, bytes, bytearray]):
        self._filename = "FromBytes.ism"
        if isinstance(filename, bytes) or isinstance(filename, bytearray):
            self._binaryFileContent = filename
            ismFile = io.BytesIO(filename)
        else:
            (_, self._filename) = os.path.split(filename)
            with open(filename, "rb") as f:
                self._binaryFileContent = f.read()

            ismFile = open(filename, "rb")

        version = readI6FromFile(ismFile)
        self.numberOfSamples = readI6FromFile(ismFile) + 1
        self.frequency = readF8ArrayFromFile(ismFile, self.numberOfSamples)
        self.impedance = readF8ArrayFromFile(ismFile, self.numberOfSamples)
        self.phase = readF8ArrayFromFile(ismFile, self.numberOfSamples)
        self.measurementTimeStamp = readTimeStampDateTimeArrayFromFile(
            ismFile, self.numberOfSamples
        )
        self.significance = readI2ArrayFromFile(ismFile, self.numberOfSamples)

        self.acqChannels = dict()

        self.measurementDate = readZahnerDate(ismFile)

        self.system = readZahnerStringFromFile(ismFile)
        self.potential = readZahnerStringFromFile(ismFile)
        self.current = readZahnerStringFromFile(ismFile)
        self.temperature = readZahnerStringFromFile(ismFile)
        self.time = readZahnerStringFromFile(ismFile)
        self.comment_1 = readZahnerStringFromFile(ismFile)
        self.comment_2 = readZahnerStringFromFile(ismFile)
        self.comment_3 = readZahnerStringFromFile(ismFile)
        self.comment_4 = readZahnerStringFromFile(ismFile)

        self.areaForCurrentDensity = readZahnerStringFromFile(ismFile)
        serialQuantityStuff = readZahnerStringFromFile(ismFile)
        acquisition_flag = readI2FromFile(ismFile)

        kValues = readF8ArrayFromFile(ismFile, 32)

        k_value_27 = int(kValues[27])

        if acquisition_flag > 256 and (k_value_27 & 32768) == 32768:
            self.acqChannels["Voltage/V"] = np.ndarray(
                shape=(self.numberOfSamples,), dtype=">f8"
            )
            self.acqChannels["Current/A"] = np.ndarray(
                shape=(self.numberOfSamples,), dtype=">f8"
            )

            for index in range(self.numberOfSamples):
                self.acqChannels["Voltage/V"][index] = readF8FromFile(ismFile)
                self.acqChannels["Current/A"][index] = readF8FromFile(ismFile)

        self._metaData = bytearray(ismFile.read())
        ismFile.close()

        """
        The frequency range at the beginning, which is measured overlapping, is not returned,
        therefore now the array indices are determined in which range the data are, which are
        returned by the getters.
        """
        self.minFrequencyIndex = np.argmin(self.frequency)
        self.maxFrequencyIndex = np.argmax(self.frequency)

        self.swapNecessary = False
        self.fromIndex = self.minFrequencyIndex
        self.toIndex = self.maxFrequencyIndex

        if self.minFrequencyIndex > self.maxFrequencyIndex:
            self.swapNecessary = True
            self.fromIndex = self.maxFrequencyIndex
            self.toIndex = self.minFrequencyIndex

        self.toIndex += 1
        return

    def getNumberOfSamples(self) -> int:
        """
        Returns the complete number of samples.

        This function returns the number of samples in the ism file.

        :returns: Number of total samples.
        """
        return self.numberOfSamples

    def _arraySlice(
        self, array: np.ndarray, includeDoubleFrequencies: bool = False
    ) -> np.ndarray:
        """
        Can return the array range without duplicate frequency support points.

        For numpy and other libraries there must not be any duplicate frequency support points, therefore this function trims the array range.

        :param array: Array that is processed.
        :param includeDoubleFrequencies: If True, all measurement data are returned, if False, only the largest non-overlapping area is returned. Defaults to False.
        :return: The eventually trimmed array.
        """
        retval = array
        if includeDoubleFrequencies is False:
            retval = retval[self.fromIndex : self.toIndex]

        if self.swapNecessary:
            return np.flip(retval)
        else:
            return retval

    def getFrequencyArray(self, includeDoubleFrequencies: bool = False) -> np.ndarray:
        """
        Get the frequency points from the measurement.

        :param includeDoubleFrequencies: If True, all measurement data are returned, if False, only the largest non-overlapping area is returned. Defaults to False.
        :returns: Numpy array with the frequency points.
        """
        return self._arraySlice(self.frequency, includeDoubleFrequencies)

    def getImpedanceArray(self, includeDoubleFrequencies: bool = False) -> np.ndarray:
        """
        Get the impedance points from the measurement.

        :param includeDoubleFrequencies: If True, all measurement data are returned, if False, only the largest non-overlapping area is returned. Defaults to False.
        :returns: Numpy array with the impedance points.
        """
        return self._arraySlice(self.impedance, includeDoubleFrequencies)

    def getPhaseArray(
        self, degree: bool = False, includeDoubleFrequencies: bool = False
    ) -> np.ndarray:
        """
        Get the phase points from the measurement.

        :param degree: True for phase in degree, default radiant.
        :param includeDoubleFrequencies: If True, all measurement data are returned, if False, only the largest non-overlapping area is returned. Defaults to False.
        :returns: Numpy array with the phase points as radiant.
        """
        radToDegree = 1.0
        if degree == True:
            radToDegree = 360 / (2 * np.pi)

        return self._arraySlice(self.phase, includeDoubleFrequencies) * radToDegree

    def getComplexImpedanceArray(
        self, includeDoubleFrequencies: bool = False
    ) -> np.ndarray:
        """
        Get the complex impedance points from the measurement.

        :param includeDoubleFrequencies: If True, all measurement data are returned, if False, only the largest non-overlapping area is returned. Defaults to False.
        :returns: Numpy array with the complex impedance points.
        """
        imp = self.getImpedanceArray(includeDoubleFrequencies)
        phase = self.getPhaseArray(includeDoubleFrequencies=includeDoubleFrequencies)
        return np.cos(phase) * imp + 1j * np.sin(phase) * imp

    def getSignificanceArray(
        self, includeDoubleFrequencies: bool = False
    ) -> np.ndarray:
        """
        Get the significance points from the measurement.

        :param includeDoubleFrequencies: If True, all measurement data are returned, if False, only the largest non-overlapping area is returned. Defaults to False.
        :returns: Numpy array with the significance points.
        """
        return self._arraySlice(self.significance, includeDoubleFrequencies)

    def getMeasurementDateTimeArray(
        self, includeDoubleFrequencies: bool = False
    ) -> np.ndarray:
        """
        Get the timestamps from the measurement.

        :param includeDoubleFrequencies: If True, all measurement data are returned, if False, only the largest non-overlapping area is returned. Defaults to False.
        :returns: Numpy array with the datetime objects.
        """
        return self._arraySlice(self.measurementTimeStamp, includeDoubleFrequencies)

    def getMeasurementStartDateTime(self) -> datetime.datetime:
        """
        Get the start date time of the measurement.

        Returns the start datetime of the measurement.

        :returns: datetime object with the start time of the measurement.
        """
        return min(self.measurementTimeStamp)

    def getMeasurementEndDateTime(self) -> datetime.datetime:
        """
        Get the end date time of the measurement.

        Returns the end datetime of the measurement.

        :returns: datetime object with the end time of the measurement.
        """
        return max(self.measurementTimeStamp)

    def save(self, filename):
        """
        Save the impedance data.

        Only the binary file content that has been read is saved. If the data is edited, this is not saved.

        :param filename: Path and filename of the file to be saved with the extension .ism.
        """
        with open(filename, "wb") as f:
            f.write(self._binaryFileContent)
        return

    def getFileName(self) -> str:
        """
        Get the name of the file.

        :returns: The filename if the file was opened or "FromBytes.ism" if it was created from bytearrays.
        """
        return self._filename

    def getBinaryFileContent(self) -> bytearray:
        """
        Get the content of the file binary.

        Returns the file contents as a binary byte array.

        :returns: bytearray with the file content.
        """
        return self._binaryFileContent

    def getMetaData(self) -> bytearray:
        """
        Get the meta data of the file binary.

        Returns the file contents as a binary byte array.

        :returns: bytearray with the file content.
        """
        return self._metaData

    def getTrackTypesList(self) -> list[str]:
        """
        returns a list with the different data tracks.

        :returns: List with the track names.
        """
        return list(map(str, self.acqChannels.keys()))

    def getTrack(
        self, track: str, includeDoubleFrequencies: bool = False
    ) -> np.ndarray:
        """
        returns an array with the points for the given track.

        :param track: name of the track.
        :param includeDoubleFrequencies: If True, all measurement data are returned, if False, only the largest non-overlapping area is returned. Defaults to False.
        :returns: Numpy array with the track.
        """
        return self._arraySlice(self.acqChannels[track], includeDoubleFrequencies)



"""
  ____       __                        __    __   __      _ __
 /_  / ___ _/ /  ___  ___ ___________ / /__ / /__/ /_____(_) /__
  / /_/ _ `/ _ \/ _ \/ -_) __/___/ -_) / -_)  '_/ __/ __/ /  '_/
 /___/\_,_/_//_/_//_/\__/_/      \__/_/\__/_/\_\\__/_/ /_/_/\_\

Copyright 2023 Zahner-Elektrik GmbH & Co. KG

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import requests
import subprocess
import os
import logging
import time


class AnalysisConnection:
    """Class which manages the connection to the Zahner Analysis Software.

    The Zahner Analysis Software performs all calculations.
    The Python library communicates with the Zahner Analysis via a REST interface.

    Various optional settings are possible. In the standard configuration it is checked if a
    Zahner Analysis with server is running, if this is not the case the Zahner Analysis is automatically
    started and automatically closed again.

    :param ip: IP address under which the server is found. 127.0.0.1 if it is running on the same computer.
    :param port: Port of the server.
    :param tryToConnect: True to test if the Zahner Analysis Server is reachable.
    :param tryToStart: If tryToConnect is True and tryToStart is True, then the Zahner Analysis is tried to start under windows systems.
    :param onlineCheckUrl: This URL is used to check if the Zahner Analysis can be reached.
    :param apiKey: The key is used as a security feature to prevent others from accessing the Zahner Analysis.
    :param buffer_size: Optimization parameter: The buffer for incomming file uploads. If too small the server has to reallocate bigger amounts of memory more often. Bigger amount means more memory allocation even if not needed.
    :param keep_jobs: Number of finished fitting jobs to keep for the client to download results and check the status.
    :param timeToWaitForOnline: This is the time for which is waited until the Zahner Analysis is online after it has been started. This time varies depending on the speed of the computer.
    :param keepOpen: If this parameter is True, the Zahner Analysis will not be closed when it is no longer needed.
    """

    def __init__(
        self,
        ip="127.0.0.1",
        port=8081,
        tryToConnect=True,
        tryToStart=True,
        onlineCheckUrl="/id",
        apiKey="PyAnalysis",
        buffer_size=32768,
        keep_jobs=3,
        timeToWaitForOnline=10,
        keepOpen=False,
    ):
        self._ip = ip
        self._port = port
        self._ip = f"http://{self._ip}:{self._port}"
        self._onlineCheckUrl = onlineCheckUrl
        self._apiKey = apiKey
        self._subprocess = None
        self._keepOpen = keepOpen

        if tryToConnect:
            if self.isOnline() is False:
                if tryToStart:
                    self.tryToStartAnalysis(buffer_size, keep_jobs, timeToWaitForOnline)
        return

    def tryToStartAnalysis(self, buffer_size=32768, keep_jobs=3, timeToWaitForCheck=10):
        """Function which tries to start the Zahner Analysis.

        This function writes information with the logging module.

        :param buffer_size: Optimization parameter: The buffer for incomming file uploads. If too small the server has to reallocate bigger amounts of memory more often. Bigger amount means more memory allocation even if not needed.
        :param keep_jobs: Number of finished fitting jobs to keep for the client to download results and check the status.
        :param timeToWaitForOnline: This is the time for which is waited until the Zahner Analysis is online after it has been started. This time varies depending on the speed of the computer.
        """
        opened = False
        if os.name == "nt":
            # Try to start Zahner Analysis only on Windows.
            opened = True
            try:
                try:
                    logging.debug("Zahner Analysis 64 bit registry search")
                    import winreg

                    reg_key = winreg.OpenKey(
                        winreg.HKEY_LOCAL_MACHINE,
                        "SOFTWARE\\Zahner\\Analysis",
                        0,
                        winreg.KEY_READ,
                    )
                except:
                    logging.debug("Zahner Analysis 32 bit registry search")
                    reg_key = winreg.OpenKey(
                        winreg.HKEY_LOCAL_MACHINE,
                        "SOFTWARE\\Zahner\\Analysis",
                        0,
                        winreg.KEY_READ | winreg.KEY_WOW64_32KEY,
                    )
                finally:
                    path = winreg.QueryValueEx(reg_key, "Path")[0]
                    reg_key.Close()

                path += " --remoteEval"
                path += f",port={self._port}"
                path += f",key={self._apiKey}"
                path += f",buffer_size={buffer_size}"
                path += f",keep_jobs={keep_jobs}"
                self._subprocess = subprocess.Popen(path)
                time.sleep(1)  # wait one second default
            except:
                opened = False

        if opened:
            logging.info("Zahner Analysis launched")
            startTime = time.time()
            # Wait up to timeToWaitForCheck seconds until the server is online.
            while (
                timeToWaitForCheck > (time.time() - startTime)
                and self.isOnline() is False
            ):
                logging.info("Zahner Analysis server offline - retry")
                time.sleep(0.5)
            if self.isOnline() is False:
                logging.error("Zahner Analysis launched but server is offline")
        else:
            logging.error("Zahner Analysis can not be launched")
        return opened

    def isOnline(self, timeout=0.5):
        """Returns the online state

        This function returns true if the Zahner Analysis can be reached.
        This function writes information with the logging module.

        :param timeout: Timeout, for the get command.
        :returns: True if the Zahner Analysis is online.
        """
        retval = True
        try:
            resp = self.get(self._onlineCheckUrl, timeout=timeout)
            if resp.json()["license-status"] == "invalid":
                retval = False
                logging.info("Zahner Analysis server online - license invalid")
            else:
                logging.info("Zahner Analysis server online")
        except:
            logging.info("Zahner Analysis server offline")
            retval = False
        return retval

    def get(self, url=None, *args, **kwargs):
        """get request wrapper

        This function wraps the get request. Only the relative URL path to the root must be specified.
        This function automatically adds ip, port and api key.

        If a full path with http is specified, then this path will be used and will not be edited.

        :param url: url for request.
        :returns: Reply from requests.get
        """
        return requests.get(self._urlComposer(url), *args, **kwargs)

    def post(self, url=None, *args, **kwargs):
        """post request wrapper

        This function wraps the post request. Only the relative URL path to the root must be specified.
        This function automatically adds ip, port and api key.

        If a full path with http is specified, then this path will be used and will not be edited.

        :param url: URL for request.
        :returns: Reply from requests.post
        """
        return requests.post(self._urlComposer(url), *args, **kwargs)

    def _urlComposer(self, url):
        """Composition of the URL

        This function composes the URL with IP, port and api key if there is no "http" in the url parameter.

        :param url: URL relative or absolute.
        :returns: Composed URL.
        """
        if url is None:
            url = self._ip
        elif "http" in url:
            pass
        else:
            url = f"{self._ip}{url}?key={self._apiKey}"
        return url

    def __del__(self):
        """Destructor

        If the library has opened the Zahner Analysis, then it will be closed again. If keepOpen is set, it will not be closed.
        """
        if self._subprocess is not None:
            if self._keepOpen is False:
                self._subprocess.kill()
        return



"""
  ____       __                        __    __   __      _ __
 /_  / ___ _/ /  ___  ___ ___________ / /__ / /__/ /_____(_) /__
  / /_/ _ `/ _ \/ _ \/ -_) __/___/ -_) / -_)  '_/ __/ __/ /  '_/
 /___/\_,_/_//_/_//_/\__/_/      \__/_/\__/_/\_\\__/_/ /_/_/\_\

Copyright 2023 Zahner-Elektrik GmbH & Co. KG

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import numpy as np
from zahner_analysis.file_import.ism_import import IsmImport
from typing import Optional


class IsmExport:
    """
    Class for saving ism data.

    With this class only impedances to different frequency points can be stored.

    However, the ism files are not complete and can no longer be loaded in Thales, nor do they contain any meta data.
    ACQ data is also lost.

    :param frequency: Array with the frequency data.
    :param impedance: Array with the impedance data.
    :param phase: Array with the phase data.
    :param metaData: Binary metadata.
    """

    def __init__(
        self,
        ism: Optional[IsmImport] = None,
        frequency: Optional[list[float]] = None,
        impedance: Optional[list[float]] = None,
        phase: Optional[list[float]] = None,
        metaData: bytearray = bytearray(),
    ):
        self.startOfFile = bytearray(b"\x00\x00\xff\xff\xff\xfe")
        self._binaryFileContent = bytearray()

        if ism is not None:
            frequency = ism.getFrequencyArray()
            impedance = np.abs(ism.getComplexImpedanceArray())
            phase = np.angle(ism.getComplexImpedanceArray())
            metaData = ism.getMetaData()

        self.numberOfElements = min([len(frequency), len(impedance), len(phase)])

        significance = np.ones(shape=(self.numberOfElements,), dtype=">i2") * 1000
        time = [i for i in range(self.numberOfElements)]

        self.tmpFrequency = np.array(frequency, dtype=">f8")
        self.tmpImpedance = np.array(impedance, dtype=">f8")
        self.tmpPhase = np.array(phase, dtype=">f8")
        self.tmpTime = np.array(time, dtype=">f8")
        self.tmpSig = np.array(significance, dtype=">i2")

        self._binaryFileContent += self.startOfFile
        numberToWrite = self.numberOfElements - 1
        self._binaryFileContent += numberToWrite.to_bytes(6, "big")
        self._binaryFileContent += self.tmpFrequency.tobytes()
        self._binaryFileContent += self.tmpImpedance.tobytes()
        self._binaryFileContent += self.tmpPhase.tobytes()
        self._binaryFileContent += self.tmpTime.tobytes()
        self._binaryFileContent += self.tmpSig.tobytes()
        self._binaryFileContent += metaData
        return

    def getBinaryFileContent(self):
        """
        Get the content of the file binary.

        Returns the file contents as a binary byte array.

        :returns: bytearray with the file content.
        """
        return self._binaryFileContent

    def writeToFile(self, file):
        """
        Writing the file to the hard disk.

        :param file: Path to the file to be written.
        """
        with open(file, "wb") as file:
            file.write(self._binaryFileContent)
        return



"""
  ____       __                        __    __   __      _ __
 /_  / ___ _/ /  ___  ___ ___________ / /__ / /__/ /_____(_) /__
  / /_/ _ `/ _ \/ _ \/ -_) __/___/ -_) / -_)  '_/ __/ __/ /  '_/
 /___/\_,_/_//_/_//_/\__/_/      \__/_/\__/_/\_\\__/_/ /_/_/\_\

Copyright 2023 Zahner-Elektrik GmbH & Co. KG

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import io
import os
import re
from io import StringIO
from typing import Union

from zahner_analysis.file_import.thales_file_utils import *


class SeqTxtImport:
    """Class to be able to read the sequence text files.

    The following lines show a short example how to read all data tracks including ACQ channels.

    .. code-block:: python

        imp = SeqTxtImport(r"C:/THALES/sequence.txt")

        tracks = imp.getTrackTypesList()
        dataTracks = {}
        for track in tracks:
            dataTracks[track] = imp.getTrack(track)

        time = imp.getTrack("Time/s")
        u = imp.getTrack("E/V")
        i = imp.getTrack("I/A")

    :param file: The path to the txt file, or the txt file as bytes or bytearray.
    :type file: str, bytes, bytearray
    """

    def __init__(self, filename: Union[str, bytes, bytearray]):
        self._filename = "FromBytes.txt"
        if isinstance(filename, bytes) or isinstance(filename, bytearray):
            self._binaryFileContent = filename
            fileBinary = io.BytesIO(filename)
        else:
            (_, self._filename) = os.path.split(filename)
            with open(filename, "rb") as f:
                self._binaryFileContent = f.read()

            fileBinary = open(filename, "rb")

        fileString = fileBinary.read().decode("utf-8")
        fileStringIO = StringIO(fileString)
        header = fileStringIO.readline()

        singlePattern = r"[\s]+(\S+)"
        self._dataTrackNames = re.findall(singlePattern, header)

        completePattern = singlePattern * len(self._dataTrackNames) + r"[\s]+"

        self._data = dict()
        for key in self._dataTrackNames:
            self._data[key] = []

        for line in fileStringIO.readlines():
            match = re.search(completePattern, line)

            for i in range(len(self._dataTrackNames)):
                self._data[self._dataTrackNames[i]].append(float(match[i + 1]))

        return

    def getTrackTypesList(self) -> list[str]:
        """
        returns a list with the different data tracks.

        :returns: List with the track names.
        """
        return self._dataTrackNames

    def getTrack(self, track: str) -> list[float]:
        """
        returns an array with the points for the given track.

        :param track: name of the track
        :returns: Numpy array with the track.
        """
        return self._data[track]

    def save(self, filename):
        """Save the cv data.

        Only the binary file content that has been read is saved. If the data is edited, this is not saved.

        :param filename: Path and filename of the file to be saved with the extension .txt.
        """
        with open(filename, "wb") as f:
            f.write(self._binaryFileContent)
        return

    def getFileName(self) -> str:
        """
        Get the name of the file.

        :returns: The filename if the file was opened or "FromBytes.txt" if it was created from bytearrays.
        """
        return self._filename

    def getBinaryFileContent(self) -> bytearray:
        """
        Get the content of the file binary.

        Returns the file contents as a binary byte array.

        :returns: bytearray with the file content.
        """
        return self._binaryFileContent



from zahner_analysis.analysis_tools.eis_fitting import EisFitting, EisFittingPlotter
from zahner_analysis.file_import.impedance_model_import import IsfxModelImport
from zahner_analysis.file_import.ism_import import IsmImport
import matplotlib.pyplot as plt

if __name__ == "__main__":
    impedanceCircuitModel = IsfxModelImport("rc-model.isfx")
    impedanceData = IsmImport("rc-data.ism")

    print(impedanceCircuitModel)

    fitting = EisFitting()

    fittingResult = fitting.fit(impedanceCircuitModel, impedanceData)

    print(fittingResult)

    foldername = "fit_results"
    fittingResult.save(foldername=foldername)

    (fig2, (impedanceAxis2, phaseAxis2)) = EisFittingPlotter.plotBode(fittingResult)

    impedanceAxis2.legend(
        impedanceAxis2.get_lines() + phaseAxis2.get_lines(),
        2 * ["Measured Data", "Fitted Model"],
    )

    fig2.set_size_inches(18, 10)
    plt.show()

    fig2.savefig(f"{foldername}/bode.svg")

    fittetModel = fittingResult.getFittedModel()

    print(
        f"{fittetModel['C0'].getType()}: {fittetModel['C0']['C'].getValue():1.3e} {fittetModel['C0']['C'].getUnit()}"
    )
    print(
        f"{fittetModel['R0'].getType()}: {fittetModel['R0']['R'].getValue():1.3e} {fittetModel['R0']['R'].getUnit()}"
    )
